import os.path

from PyQt5 import QtWidgets
from PyQt5.QtCore import QObject, QUrl, pyqtProperty, pyqtSignal, pyqtSlot

from cura.Settings.ContainerManager import ContainerManager
from cura.Settings.MachineManager import MachineManager
from cura.SIProxies.GencePrinterManager import GencePrinterManager
from cura.SIProxies.SettingsManage.PrintersTree import PrinterTree
from UM.Application import Application
from UM.Resources import Resources


class CustomSettingsActions(QObject):
    instance = None

    def __init__(self):
        super().__init__()
        self._loadedProfilePath = None
        self._restoreLastLoadedProfileOnClose = True
        self._containerManager = ContainerManager.getInstance()
        self._machineManager = Application.getInstance().getMachineManager()
        self._profileName = None
        self._profileExt = None

    profileNameChanged = pyqtSignal()

    @pyqtProperty(str, notify = profileNameChanged)
    def profileName(self):
        return self._profileName if self._profileName else ""

    @profileName.setter
    def profileName(self, name: str):
        self._profileName = name
        self.profileNameChanged.emit()

    @pyqtSlot()
    def activateProfile(self, name = "custom_settings"):
        self._containerManager.createQualityChanges(name, full_profile = True)

    @pyqtSlot()
    def deactivateProfile(self, name = "custom_settings"):
        self._containerManager.removeQualityChanges(name)

    @pyqtSlot()
    def exportProfileDialog(self):
        path = self._profileNameToPath()
        saveFile = QtWidgets.QFileDialog.getSaveFileName(None, 'Save profile as...', path, 'Profile (*.profile)', options = QtWidgets.QFileDialog.DontConfirmOverwrite)
        if saveFile[0] is not '':
            pathToSave = self.setPathExtension(saveFile)
            result = self.exportProfileAsFile(pathToSave, saveFile[1])
            return result
        else:
            return False
        
    def exportProfileAsFile(self, path, fileType):
        profileName = self.pathToProfileName(path)
        self.profileName = profileName
        qurlPath = QUrl().fromLocalFile(path)
        self.activateProfile(profileName)
        containers = self._containerManager.findInstanceContainers({"type": "quality_changes", "name": self._machineManager.activeQualityName})
        result = self._containerManager.exportProfile(containers, qurlPath, fileType)
        self.deactivateProfile(profileName)
        return result

    def setPathExtension(self, saveFile):
        if saveFile[1] == 'Profile (*.profile)':
            saveFilePath = saveFile[0]
            if saveFilePath[-8:] != '.profile':
                path = saveFile[0] + '.profile'
            else:
                path = saveFile[0]
        else:
            path = saveFile[0]

        return path

    @pyqtSlot(result = bool)
    def loadProfileDialog(self):
        loadFile = QtWidgets.QFileDialog.getOpenFileName(None, 'Load profile', Resources.getCustomProfilesPath(), 'Profile file (*.profile);;All files (*)')
        if loadFile[0] is not '':
            pathToLoad = loadFile[0]
            result = self.loadProfileFromFile(pathToLoad)
            return result
        else:
            return False

    def loadProfileFromFile(self, path):
        self.clearUserSettings()
        profileName = self.pathToProfileName(path)
        result = GencePrinterManager.getInstance().setProfileFromPath(path, profileName)
        if result == True:
            PrinterTree.getInstance()._customProfileLoaded = True
            self._loadedProfilePath = path
            self.profileName = self.pathToProfileName(path)
        
        return result

    @pyqtSlot()
    def loadLastProfile(self):
        self.clearUserSettings()
        gencePrinterManager = GencePrinterManager.getInstance()
        gencePrinterManager.reloadProfile()

    def pathToProfileName(self, path):
        filename = os.path.basename(path)
        profileName = os.path.splitext(filename)[0]

        self._profileExt = os.path.splitext(filename)[1]
        
        return profileName

    def _profileNameToPath(self):
        dirName = os.path.dirname(self._loadedProfilePath)
        path = os.path.join(dirName, self._profileName) + self._profileExt
        return path

    def clearUserSettings(self):
        self._containerManager.clearUserContainers()


def createCustomSettingsActions(engine, script_engine):
    if(CustomSettingsActions.instance==None):
        CustomSettingsActions.instance = CustomSettingsActions()

    return CustomSettingsActions.instance

def getInstance():
    if(CustomSettingsActions.instance==None):
        CustomSettingsActions.instance = CustomSettingsActions()

    return CustomSettingsActions.instance
