from PyQt5.QtCore import QCoreApplication, Qt, pyqtProperty, pyqtSlot

from cura.Settings.ExtrudersModel import ExtrudersModel
from cura.Settings.SettingOverrideDecorator import SettingOverrideDecorator
from cura.SIProxies.SettingsManage.PrintersTree import PrinterTree
from UM.Application import Application
from UM.Operations.OperationStack import OperationStack
from UM.Qt.ListModel import ListModel
from UM.Scene.Selection import Selection


class PrintableNodesModel(ListModel):
    NodeNameRole = Qt.UserRole+1
    SelectedRole = Qt.UserRole+2
    NodeRole = Qt.UserRole+3
    ExtruderRole = Qt.UserRole+4

    def __init__(self,parent=None):
        if Application.getInstance()._printable_nodes != None:
            raise ValueError("Duplicate singleton creation of PrintableNodesModel")
        Application.getInstance()._printable_nodes = self

        super().__init__(parent)
        self._extruders_model = ExtrudersModel()

        self.addRoleName(self.NodeNameRole, "name")
        self.addRoleName(self.SelectedRole, "selected")
        self.addRoleName(self.ExtruderRole, "activeExtruder")
        Selection.selectionChanged.connect(self._update)
        operationStack = Application.getInstance().getOperationStack()
        operationStack.changed.connect(self._update)
        # SelectedPrinterChanged connect needed to handle situation where activeExtruder was changed and then printer
        # configuration is changed to single extruder profile.
        printersTree = PrinterTree.getInstance()
        printersTree.selectedPrinterChanged.connect(self._update)

    def _update(self):
        # items = []
        selectableNodes = Selection.getAllSelectableNodes()

        # Node deleted
        if len(selectableNodes) < len(self.items):
            for i, item in enumerate(self.items):
                stillExists = False
                for node in selectableNodes:
                    if item["node"] == node:
                        stillExists = True
                        break
                
                if not stillExists:
                    self.removeItem(i)
                    break

        else:
            for i, node in enumerate(selectableNodes):
                isSelected = False
                if Selection.isSelected(node):
                    isSelected = True
                
                activeExtruder = None
                if node.hasDecoration("setActiveExtruder"):
                    activeExtruder = node.callDecoration("getActiveExtruderPosition")

                current_item = {"name": node.getName(), "selected": isSelected, "node": node, "activeExtruder": activeExtruder}
                existing_item = self.getItem(i)

                if existing_item == {}:
                    self.appendItem(current_item)
                elif existing_item != current_item:
                    self.insertItem(i, current_item)
                    self.removeItem(i+1)
                    
                # items.append(item)

        # self.setItems(items)

    @classmethod
    def getInstance(cls):
        if  Application.getInstance()._printable_nodes is None:
            Application.getInstance()._printable_nodes = cls()

        return Application.getInstance()._printable_nodes

    @pyqtSlot(int)
    def handleSingleSelection(self, index):
        currently_selected = self._countSelected()
        is_item_selected = self._isItemSelected(index)

        if currently_selected == 1 and is_item_selected:
            pass
        elif currently_selected != 0:
            self._clearSelection()

        self.toggleSelection(index)

    @pyqtSlot(int)
    def toggleSelection(self, index):
        current_state = self.getItem(index)["selected"]
        new_state = not current_state
        node = self.getItem(index)["node"]
        # self.setProperty(index, "selected", new_state)
        if new_state:
            Selection.add(node)
        else:
            Selection.remove(node)
            
    @pyqtSlot(int)
    def toggleExtruder(self, index):
        current_state = self.getItem(index)["activeExtruder"]
        node = self.getItem(index)["node"]
        new_state = None

        if current_state == None:
            node.addDecorator(SettingOverrideDecorator())
            new_state = 0
        elif current_state == '0':
            new_state = 1
        elif current_state == '1':
            node.removeDecorator(SettingOverrideDecorator)

        if new_state != None:
            extruder_stack_id = self._extruders_model.getItem(new_state)["id"]
            node.callDecoration("setActiveExtruder", extruder_stack_id)
        self._update()

    def _countSelected(self):
        result = 0
        for item in self._items:
            if item["selected"]:
                result += 1

        return result

    def _clearSelection(self):
        Selection.clear()
        # for item_id in range(len(self._items)):
            # self.setProperty(item_id, "selected", False)

    def _isItemSelected(self, index):
        return self._items[index]["selected"]

def CreatePrintableNodesModel(engine, script_engine):
    return PrintableNodesModel.getInstance()
