from math import pi as mathpi
from UM.Tool import Tool
from UM.Event import Event, MouseEvent
from UM.Math.Vector import Vector
from UM.Application import Application
from UM.Math.Quaternion import Quaternion

from . import SelectionPlaneNode
from . import GeometryToolbox as GeometryToolbox
from .PolygonSelectionPass import PolygonSelectionPass
from .GenceOrientFaceToPlaneOperation import GenceOrientFaceToPlaneOperation

from cura.SIProxies.NotificationsProxy import NotificationsProxy

class GenceOrientFaceToPlaneTool(Tool):
    def __init__(self):
        super().__init__()
        self._controller = Application.getInstance().getController()
        self._scene = Application.getInstance().getController().getScene()
        self._render = Application.getInstance().getRenderer()

        self._selection_plane = SelectionPlaneNode.SelectionPlaneNode()

        self._polygon_selection_pass = None
        self._selection_pass = None
        self._current_node_id = None

        self._meshes_cache = dict()

    def event(self, event):
        super().event(event)

        if not self._polygon_selection_pass:
            self._render.addRenderPass(PolygonSelectionPass())
            self._polygon_selection_pass = self._render.getRenderPass("polygon_selection")

        if not self._selection_pass:
            self._selection_pass = self._render.getRenderPass("selection")

        if event.type == Event.MouseMoveEvent:
            self._calculate_selection_plane(event)

        if event.type == Event.MousePressEvent:
            if MouseEvent.LeftButton not in event.buttons:
                return False

            polygon_information = self._getPolygonColor(event)
            if polygon_information is not None:
                self._calculate_model_rotation(polygon_information)
            else:
                NotificationsProxy.createNotificationsProxy().deactivatePlaceOnBedTool()

        if event.type == Event.ToolActivateEvent:
            self._controller.activateSelectionTool(False)

        if event.type == Event.ToolDeactivateEvent:
            self._controller.activateSelectionTool(True)
            self._showTool(None)
            self._polygon_selection_pass.setVisibleNode(None)
            NotificationsProxy.createNotificationsProxy().deactivatePlaceOnBedTool(False)
            self._meshes_cache = {}

    def _calculate_selection_plane(self, event):
        polygon_info = self._getPolygonColor(event)
        polygon = self._convertColorToPolygon(polygon_info)
        self._showTool(polygon)

    def _calculate_model_rotation(self, polygon_information):
        polygon = self._convertColorToPolygon(polygon_information, False)

        if polygon:
            self._placeOnPolygon(polygon)

    def _showTool(self, polygon):
        if polygon:
            self._selection_plane.showSelectionPlane(polygon["polygon_verticies"])
        else:
            self._selection_plane.hideSelectionPlane()

    def _getPolygonColor(self, event):
        node_id = self._selection_pass.getIdAtPosition(event.x, event.y)

        if node_id != self._current_node_id:
            self._current_node_id = node_id
            self._polygon_selection_pass.setVisibleNode(node_id)
            self._scene.sceneChanged.emit(self)

        if node_id is not None:
            color = self._polygon_selection_pass.getColorAtPosition(event.x, event.y)
            return {'node_id': node_id, 'polygon_color': color}

        return None

    # Polygon colors are assigned in MeshData._init
    def _convertColorToPolygon(self, pol_info, transformed = True):
        if pol_info is None:
            return None

        node_id = pol_info['node_id']
        polygon_color = (pol_info['polygon_color'] & 0x00ffffff)

        node = self._scene.findObject(node_id)
        if node:
            if transformed:
                mesh = self._meshes_cache.get(node_id, None)
                if mesh is None:
                    self._meshes_cache[node_id] = mesh = node.getMeshDataTransformed()
            else:
                mesh = node.getMeshData()

            vertices = mesh.getVertices()

            base = self._calculateFaceIndex(mesh.getVertexCount(), polygon_color) * 3

            result = dict()
            result["node_id"] = node_id
            result["first_vertex_index"] = base
            result["polygon_verticies"] = (vertices[base], vertices[base+1], vertices[base+2])

            return result

        return None

    def _calculateFaceIndex(self, vertex_count, polygon_color):
        color_count = vertex_count / 3
        step = int((256 ** 3) / color_count)
        face_num = polygon_color / step

        return int(face_num)

    def _placeOnPolygon(self, polygon_info):
        node_id = polygon_info['node_id']
        polygon = polygon_info['polygon_verticies']

        node = self._scene.findObject(node_id)
        vertex_0, vertex_1, vertex_2 = Vector(data=polygon[0]), Vector(data=polygon[1]), Vector(data=polygon[2])

        p_norm = GeometryToolbox.CalculatePolygonNormal(vertex_0, vertex_1, vertex_2)

        rotation_matrix = GeometryToolbox.CalculateRotationToParallel(p_norm, Vector(0, -1.0, 0))

        node.setOrientation(Quaternion.fromMatrix(rotation_matrix))

        t_mesh = node.getMeshDataTransformed()
        vertices = t_mesh.getVertices()

        vertex_0, vertex_1, vertex_2 = vertices[polygon_info["first_vertex_index"]],\
                                       vertices[polygon_info["first_vertex_index"] + 1],\
                                       vertices[polygon_info["first_vertex_index"] + 2]

        vertex_0 = Vector(x=vertex_0[0], y=vertex_0[1], z=vertex_0[2])
        vertex_1 = Vector(x=vertex_1[0], y=vertex_1[1], z=vertex_1[2])
        vertex_2 = Vector(x=vertex_2[0], y=vertex_2[1], z=vertex_2[2])

        p_norm = GeometryToolbox.CalculatePolygonNormal(vertex_0, vertex_1, vertex_2)

        q_rotation = Quaternion.fromMatrix(rotation_matrix)

        if 0.0 < p_norm.y:
            q2 = Quaternion.fromAngleAxis(mathpi, Vector.Unit_X)
            q_rotation = q2 * q_rotation

        node.setOrientation(Quaternion())

        GenceOrientFaceToPlaneOperation(node, q_rotation).push()
        self.setNodeAtZeroLevel(node)
        self._polygon_selection_pass.setVisibleNode(id(node))

        if self._meshes_cache.get(node_id, None):
            del self._meshes_cache[node_id]


    def invalidateMeshCache(self):
        self._meshes_cache = {}

        #Application.getInstance().getController().getTool('GenceRotateTool').rotateAndSet(node, q_rotation, x, y, z)
