from UM.Scene.SceneNode import SceneNode
from UM.Mesh.MeshBuilder import MeshBuilder
from UM.Math.Vector import Vector
from UM.Math.Color import Color
from UM.Math.Float import Float
from UM.Math.Quaternion import Quaternion
from UM.Application import Application
from UM.Resources import Resources
from UM.View.GL.OpenGL import OpenGL

from math import pi as mathpi

from . import GeometryToolbox as gt

class SelectionPlaneNode(SceneNode):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._calculate_aabb = False
        self._last_polygon = None
        self._scene = Application.getInstance().getController().getScene()

        self.setParent(self._scene.getRoot())

        self._shader = None

        self._platform_color = Color.fromHexString("#B4FF1010")
        self._bold_color = Color.fromHexString("#7A2929")

    def showSelectionPlane(self, polygon):
        if self._last_polygon is None or not self._samePolygons(polygon, self._last_polygon) or not self.isVisible():
            self.setVisible(True)
            self._last_polygon = polygon
            self._calculateMesh(polygon)

    def hideSelectionPlane(self):
        if self.isVisible() is True:
            self.setVisible(False)
            self.meshDataChanged.emit(self)

    def _samePolygons(self, a, b):
        if a is None or b is None:
            return False

        if (a[0] != b[0]).any():
            return False
        if (a[1] != b[1]).any():
            return False
        if (a[2] != b[2]).any():
            return False

        return True

    def _calculateMesh(self, polygon):
        vertex_1, vertex_2, vertex_3 = Vector(data=polygon[0]), Vector(data=polygon[1]), Vector(data=polygon[2])

        polygon_normal = gt.CalculatePolygonNormal(vertex_1, vertex_2, vertex_3)
        polygon_centroid = gt.CalculateCentroid(vertex_1, vertex_2, vertex_3)
        rotation_matrix = gt.CalculateRotationToParallel(Vector.Unit_Y, polygon_normal)

        if self.getMeshData() is None:
            self._buidBasicPlaneMesh()

        q_rotation = Quaternion.fromMatrix(rotation_matrix)
        self.setOrientation(q_rotation)

        #check if there is a need to rotate about additional 180 deg
        mesh = self.getMeshDataTransformed()
        verticies0 = mesh.getVertex(0)
        verticies1 = mesh.getVertex(1)
        verticies2 = mesh.getVertex(2)
        vertex_1, vertex_2, vertex_3 = Vector(data=verticies0), Vector(data=verticies1), Vector(data=verticies2)
        plat_normal = gt.CalculatePolygonNormal(vertex_1, vertex_2, vertex_3)

        if Float.fuzzyCompare(plat_normal.dot(polygon_normal), -1.0, 1e-6):
            q2 = Quaternion.fromAngleAxis(mathpi, Vector.Unit_X)
            q_rotation = q_rotation * q2
            self.setOrientation(Quaternion())
            self.setOrientation(q_rotation)

        self.setPosition(position=polygon_centroid + polygon_normal)

    def _buidBasicPlaneMesh(self):
        mb_square = MeshBuilder()

        # Plane
        mb_square.addQuad(Vector(0, 0, 0), Vector(40, 0, 0), Vector(40, 0, 40), Vector(0, 0, 40), color=self._platform_color)

        # Plane axis
        mb_square.addCube(0.7, 10, 0.7, center=Vector(20, 5, 20), color=self._bold_color)

        # Border
        mb_square.addCube(0.5, 0.5, 40, Vector(0,0,20), color=self._bold_color)
        mb_square.addCube(0.5, 0.5, 40, Vector(40, 0, 20), color=self._bold_color)
        mb_square.addCube(40.5, 0.5, 0.5, Vector(20, 0, 0), color=self._bold_color)
        mb_square.addCube(40.5, 0.5, 0.5, Vector(20, 0, 40), color=self._bold_color)

        # Arrow
        mb_square.addPyramid(2, 3, 2, color=self._bold_color, center = Vector(20, 10, 20))

        self.setMeshData(mb_square.build())
        self.setCenterPosition(Vector(20, 0, 20))

    def render(self, renderer):
        if self.isVisible():
            mesh = self.getMeshData()

            if mesh:
                if not self._shader:
                    self._shader = OpenGL.getInstance().createShaderProgram(Resources.getPath(Resources.Shaders, "default.shader"))

                #Auto scale
                camera_position = self._scene.getActiveCamera().getWorldPosition()
                dist = (camera_position - self.getWorldPosition()).length()
                scale = dist / 400
                self.setScale(Vector(scale, scale, scale))

                renderer.queueNode(self, mesh=mesh, overlay=True,  shader=self._shader)

        return True
