import math
import numpy as np
from scipy.ndimage import binary_erosion, binary_dilation, binary_fill_holes

from .Split import Side, Axis
from .SplitMesh import SplitMesh

from UM.Application import Application
from UM.Scene.Selection import Selection


SPLIT_DEBUG = False
PRECISION = 4


def AdvancedSplitMesh(axis, value, edge, angle, offset, length, invert):
    print("SimpleSplitMesh {{axis : {}, cut_value : {}, edge : {}, angle : {}, offset : {}, length : {}, invert : {}}}"
          .format(str(axis), str(value), str(edge), str(angle), str(offset), str(length), str(invert)))
    job = AdvancedSplitMeshJob(axis, value, edge, angle, offset, length, invert)
    job.run() if SPLIT_DEBUG else job.start()


class AdvancedSplitMeshError(Exception):
    def __init__(self, msg=None):
        if msg is None:
            msg = "Error while splitting"
        super().__init__(msg)


class AdvancedSplitMeshJob(SplitMesh):
    def __init__(self, axis, value, edge, angle, offset, length, invert):
        value = np.float32(value)
        super().__init__(axis, value, SPLIT_DEBUG)

        # Make parameters available
        self.__axis = axis
        self.__value = value
        self.__edge = np.float32(edge)
        self.__angle = int(round(angle))
        self.__offset = np.float32(offset)
        self.__length = np.float32(length)
        self.__invert = invert

        # Used data
        self.__joint_matrix = None

        self.__m_height = None
        self.__m_width = None

        self.__max_top = None
        self.__max_bottom = None
        self.__max_left = None
        self.__max_right = None

        self.__holes = []
        self.__first_shape_segments = []

    def run(self):
        success = False
        self._initialize_progress(1, "Splitting...")

        node = Selection.getCurrentlySelectedObjects()[0]
        mesh = node.getMeshDataTransformed()
        vertices = mesh.getVertices()
        normals = mesh.getNormals()

        used_polygons = []
        used_normals = []

        #TODO: reagować jak nie ma sensu dzielić na cokolwiek
        cut_halfs = self._calculate_cut_halfs(vertices, normals, 0.2)

        # Create list of verticies with at least one point in area
        # we will use
        #TODO: Można to robić w trakcie _calculate_cut_halfs(), sprawidzić czas na czymś dużym eg. yoda
        for i, vert_list in enumerate(self._grouped(vertices, 3)):
            if self.isInRange(vert_list):
                used_polygons.append(vert_list)
                used_normals.append(normals[i*3])

        try:
            # Find shapes on every layer
            shapes_in_layer = []

            special_length = False
            iterations = int(round(self.__length * 4) / 4) + 1

            if int(round(self.__length * 4)) != int(self.__length * 4):
                special_length = True

            for i in range(iterations):
                i = i if not self.__invert else -i
                self.setSplitMeshValue(self.__value - i)
                shapes = self.findShapes(used_polygons, used_normals)
                if shapes:
                    shapes_in_layer.append(shapes)

            # Special case when provided length is not integer
            if special_length:
                length = self.__length if not self.__invert else -self.__length
                self.setSplitMeshValue(self.__value - length)
                shapes = self.findShapes(used_polygons, used_normals)
                if shapes:
                    shapes_in_layer.append(shapes)

            self.setSplitMeshValue(self.__value)

            new_verts_left, new_verts_right = self.__calculate_joint(shapes_in_layer)

            # Small adjustments
            if self.__axis == Axis.X or self.__axis == Axis.Z:
                if new_verts_right:
                    new_verts_right.reverse()
            if self.__axis == Axis.Y:
                if new_verts_left:
                    new_verts_left.reverse()

            self._add_node_progress(0.1)

            cut_halfs['left_vertices'].extend(new_verts_left)
            cut_halfs['right_vertices'].extend(new_verts_right)

            self._swap_nodes_operation(node, cut_halfs['left_vertices'], cut_halfs['right_vertices'])
            self._display_status("Model split successfully")
            success = True
        except AdvancedSplitMeshError as e:
            self._display_status("Error: " + str(e))
        except Exception as e:
            self._display_status("Error: " + str(e))

        self._finish_progress(success)

    # TODO: Sprawdzić czy nie nadaje sie do użycia gdzieś indziej
    def findShapes(self, used_polygons, used_normals):
        sections_on_cut = []
        segments_normals = []

        step = int(max((len(used_polygons) / 100), 1))
        progress_part = 6 / (10 * (self.__length + 1))

        for i, polygon in enumerate(used_polygons):
            side, vertices_sides = self._checkPolygonSide(polygon)

            if side == Side.Cut:
                current_left_vertices, current_right_vertices = self._calculateSplittedPolygons(polygon, vertices_sides)
                current_left_cut_sections = self._findOnCutSections(current_left_vertices)
                sections_on_cut.extend(current_left_cut_sections)
                segments_normals += len(current_left_cut_sections) * [used_normals[i]]
            else:
                section = None
                if vertices_sides.count(Side.OnPlane) == 2:
                    section = self._findOnCutSections(polygon)

                    if section is not None and side == Side.Left:
                        sections_on_cut.extend(section)
                        segments_normals += len(section) * [used_normals[i]]

            if i % step == 0:
                self._add_node_progress((1 / 100) * progress_part)

        #self._debug_drawSections(sections_on_cut)
        if segments_normals and sections_on_cut:
            sections_normals = np.delete(segments_normals, self.__axis, 1)
            shapes = self._findShapesOnCut(sections_on_cut, sections_normals)
        else:
            raise AdvancedSplitMeshError("Cannot cut this model with those options")

        return shapes

    def isInRange(self, verticies_list):
        in_range = False

        if not self.__invert:
            left_max = self.__value - self.__length
            right_max = self.__value
        else:
            left_max = self.__value
            right_max = self.__value + self.__length


        # At least one is inside
        for vertex in verticies_list:
            if left_max < vertex[self.__axis] < right_max:
                in_range = True

        if in_range is False:
            side = None
            for vertex in verticies_list:
                if vertex[self.__axis] <= left_max:
                    if side is None or side == Side.Left:
                        side = Side.Left
                        continue
                    if side == Side.Right:
                        in_range = True
                        break
                if right_max <= vertex[self.__axis]:
                    if side is None or side == Side.Right:
                        side = Side.Right
                        continue
                    if side == Side.Left:
                        in_range = True
                        break

        return in_range

    def __calculate_joint(self, layers):
        self.__initialize_joint_data(layers)

        for layer in layers:
            np.logical_and(self.__project_shapes(layer), self.__joint_matrix, self.__joint_matrix)

        for shape in layers[0]:
            if shape['fill'] is not True:
                self.__holes.append(shape['fill'])

        return self.__get_joint_sections()

    def __initialize_joint_data(self, layers):
        for shape in layers[0]:
            self.__first_shape_segments.extend(shape['segments'])
        vertices = []

        for shapes in layers:
            for shape in shapes:
                for segment in shape['segments']:
                    vertices.append(segment[0])

        vertices = np.asarray(vertices, dtype=np.float32)

        self.__max_top = PRECISION * np.amax(vertices[:, 0])
        self.__max_bottom = PRECISION * np.amin(vertices[:, 0])
        self.__max_right = PRECISION * np.amax(vertices[:, 1])
        self.__max_left = PRECISION * np.amin(vertices[:, 1])

        self.__m_height = (math.ceil(self.__max_top) - math.floor(self.__max_bottom)) + 3
        self.__m_width = (math.ceil(self.__max_right) - math.floor(self.__max_left)) + 3

        self.__joint_matrix = np.ones(shape=(self.__m_width, self.__m_height), dtype=int)

    #TODO: Tutaj nastepuję niechciane odwrócenie?
    def __project_shapes(self, shapes):
        matrix = np.zeros(shape=(self.__m_width, self.__m_height), dtype=int)

        for shape in shapes:
            matrix_shape = np.zeros(shape=(self.__m_width, self.__m_height), dtype=int)
            shape_coords = []

            for segment in shape['segments']:
                coords = self._bresenham_line(PRECISION * segment[0][1], PRECISION * segment[0][0], PRECISION * segment[1][1], PRECISION * segment[1][0])
                shape_coords.extend(coords)

                for coord in coords:
                    # TODO: uzaleznić od normalnych ?
                    x, y = self.__calculate_coordinates(coord)

                    matrix_shape[x][y] = 1

            matrix_shape = binary_fill_holes(matrix_shape).astype(dtype=int)

            if shape['fill'] is not True:
                matrix = matrix - matrix_shape
            else:
                matrix = matrix + matrix_shape

        return matrix

    def __calculate_coordinates(self, coords):
        x = int(round(coords[0]) - math.ceil(self.__max_left) + 1)
        y = int(round(coords[1]) - math.ceil(self.__max_bottom) + 1)
        return x, y

    def __calculate_width(self, depth, angle):
        return depth/math.tan(math.radians(angle))

    def __merge_indexed(self, one_indexed, two_indexed):
        shift = len(one_indexed['vertices'])
        shifted = {}

        shifted_segments = []
        for segment in two_indexed['segments']:
            shifted_segments.append([segment[0] + shift, segment[1] + shift])

        shifted['vertices'] = []
        shifted['vertices'].extend(one_indexed['vertices'])
        shifted['vertices'].extend(two_indexed['vertices'])

        shifted['segments'] = []
        shifted['segments'].extend(one_indexed['segments'])
        shifted['segments'].extend(shifted_segments)

        shifted['merge_index'] = shift

        return shifted

    def __prepare_segments(self, segments):
        for i, segment in enumerate(segments):
            segments[i] = np.asarray([
                ((segment[0][1] + self.__max_bottom - 1) / PRECISION, (segment[0][0] + self.__max_left - 1) / PRECISION),
                ((segment[1][1] + self.__max_bottom - 1) / PRECISION, (segment[1][0] + self.__max_left - 1) / PRECISION),
            ], dtype=np.float32)

    def __get_joint_sections(self):
        width = round(PRECISION * self.__calculate_width(self.__length - self.__offset , self.__angle))
        offset = round(self.__offset * PRECISION)
        is_offset = False if offset == 0 else True

        if is_offset is False:
            new_right_verts = self.__calculate_single_joint(width)
            new_left_verts = new_right_verts.copy()
        else:
            new_right_verts = self.__calculate_single_joint(width, True, offset)

            width = round(PRECISION * self.__calculate_width(self.__length, self.__angle))
            new_left_verts = self.__calculate_single_joint(width)

        return new_left_verts, new_right_verts

    def __calculate_single_joint(self, width, offset_tenon=False, offset=0):
        erosion_inner = round(PRECISION * self.__edge) + offset + width + 1  # One to be sure we don't out bcs of matrix render

        inner_shape = self.__erosion_by_circle(erosion_inner)

        if width != 0:
            erosion_outer = round(PRECISION * self.__edge) + offset + 1
            outer_shape = self.__erosion_by_circle(erosion_outer)
            new_verts = self.__calculate_single_acute_joint_verts(inner_shape, outer_shape, offset_tenon)
        else:
            new_verts = self.__calculate_single_right_joint_verts(inner_shape, offset_tenon)

        return new_verts

    def __erosion_by_circle(self, radius):
        mask = self.__create_mask(radius)
        new_shape = binary_erosion(self.__joint_matrix, structure=mask).astype(self.__joint_matrix.dtype)
        if not new_shape.any():  # faster than numpy.count_nonzero(a) because it can return immediately when the first nonzero element has been found.
            raise AdvancedSplitMeshError("Cannot cut this model with those options")

        return new_shape

    def __create_mask(self, radius):
        diameter = 2 * radius + 1

        xx, yy = np.mgrid[:diameter, :diameter]
        ring = (xx - radius) ** 2 + (yy - radius) ** 2
        circle = np.logical_and(ring <= radius ** 2, ring >= 0)
        # self.__debug_showMatrix(circle, diameter, diameter)

        return circle

    def __calculate_single_right_joint_verts(self, shape, offset_tenon = False):
        new_verts = []

        shape_sections, _ = self.__find_shapes_in_matrix(shape)
        self.__prepare_segments(shape_sections)

        sections_indexed = self.getIndexedData(shape_sections)
        first_layer_sections_indexed = self.getIndexedData(self.__first_shape_segments)

        merged_indexed = self.__merge_indexed(sections_indexed, first_layer_sections_indexed)
        merged_indexed['holes'] = self.__holes

        result_2d = self._delaunay_constrained_triangulation(merged_indexed)

        if not offset_tenon:
            mortise = self.__value - self.__length if not self.__invert else self.__value + self.__length
        else:
            mortise = self.__value - self.__length + self.__offset if not self.__invert else self.__value + self.__length - self.__offset

        for verts in result_2d['polygons']:
            if verts[0] < merged_indexed['merge_index'] and verts[1] < merged_indexed['merge_index'] and verts[2] < merged_indexed['merge_index']:
                vert0 = np.insert(result_2d['vertices'][verts[0]], self.__axis, mortise)
                vert1 = np.insert(result_2d['vertices'][verts[1]], self.__axis, mortise)
                vert2 = np.insert(result_2d['vertices'][verts[2]], self.__axis, mortise)
                new_verts.extend([vert0, vert1, vert2])
            else:
                vert0 = np.insert(result_2d['vertices'][verts[0]], self.__axis, self.__value)
                vert1 = np.insert(result_2d['vertices'][verts[1]], self.__axis, self.__value)
                vert2 = np.insert(result_2d['vertices'][verts[2]], self.__axis, self.__value)
                new_verts.extend([vert0, vert1, vert2])

        for segment in sections_indexed['segments']:
            vert0 = np.insert(result_2d['vertices'][segment[0]], self.__axis, mortise)
            vert1 = np.insert(result_2d['vertices'][segment[0]], self.__axis, self.__value)
            vert2 = np.insert(result_2d['vertices'][segment[1]], self.__axis, mortise)
            new_verts.extend([vert0, vert1, vert2])
            vert0 = np.insert(result_2d['vertices'][segment[0]], self.__axis, self.__value)
            vert1 = np.insert(result_2d['vertices'][segment[1]], self.__axis, self.__value)
            vert2 = np.insert(result_2d['vertices'][segment[1]], self.__axis, mortise)
            new_verts.extend([vert0, vert1, vert2])

        return new_verts


    def __calculate_single_acute_joint_verts(self, inner_shape, outer_shape, offset_tenon = False):
        new_verts = []

        inner_sections, _ = self.__find_shapes_in_matrix(inner_shape)
        outer_sections, _ = self.__find_shapes_in_matrix(outer_shape)

        self.__prepare_segments(inner_sections)
        self.__prepare_segments(outer_sections)

        outer_sections.extend(self.__first_shape_segments)

        inner_indexed = self.getIndexedData(inner_sections)
        outer_indexed = self.getIndexedData(outer_sections)

        merged_indexed = self.__merge_indexed(inner_indexed, outer_indexed)
        merged_indexed['holes'] = self.__holes

        result_2d = self._delaunay_constrained_triangulation(merged_indexed)

        if not offset_tenon:
            mortise = self.__value - self.__length if not self.__invert else self.__value + self.__length
        else:
            mortise = self.__value - self.__length + self.__offset if not self.__invert else self.__value + self.__length - self.__offset


        result = []
        for i in range(0, merged_indexed['merge_index']):
            result.append(np.insert(result_2d['vertices'][i], self.__axis, mortise))

        for i in range(merged_indexed['merge_index'], len(result_2d['vertices'])):
            result.append(np.insert(result_2d['vertices'][i], self.__axis, self.__value))

        for polygon in result_2d['polygons']:
            new_verts.append(result[polygon[0]])
            new_verts.append(result[polygon[1]])
            new_verts.append(result[polygon[2]])

        return new_verts


    def __find_shapes_in_matrix(self, input_matrix=None):
        input_matrix = self.__joint_matrix if input_matrix is None else input_matrix
        inside = False

        matrix = np.zeros(shape=(self.__m_width, self.__m_height), dtype=int)

        sections = []

        # matrix - borderImage
        # self.join_matrix - paddedImage

        neighborhood = [
            [-1, 0, 7], #1
            [-1, 1, 7], #2

            [0, 1, 1], #3
            [1, 1, 1], #4

            [1, 0, 3], #5
            [1, -1, 3],#6

            [0, -1, 5],  #7
            [-1, -1, 5], #8
        ]

        for y in range(self.__m_height):
            for x in range(self.__m_width):
                if matrix[x][y] == 1 and not inside: # Entering an already discovered border
                    inside = True
                elif input_matrix[x][y] == 1 and inside: # Already discovered border point
                    continue
                elif input_matrix[x][y] == 0 and inside: # Leaving a border
                    inside = False
                elif input_matrix[x][y] == 1 and not inside: # Undiscovered border point (na bogato)
                    new_section = [(x, y)]
                    last_neighbor = None
                    last_neighbor_location_nr = None

                    matrix[x][y] = 1

                    check_location_nr = 1
                    new_check_location_nr = None

                    start_x, start_y = x, y
                    new_x, new_y = x, y

                    counter1 = 0
                    counter2 = 0

                    while True:
                        check_x = new_x + neighborhood[check_location_nr-1][0]
                        check_y = new_y + neighborhood[check_location_nr-1][1]
                        new_check_location_nr = neighborhood[check_location_nr-1][2]

                        if input_matrix[check_x][check_y] == 1:
                            if last_neighbor_location_nr is None:
                                last_neighbor_location_nr = check_location_nr
                                last_neighbor = (check_x, check_y)
                            elif last_neighbor_location_nr != check_location_nr:
                                new_section.append(last_neighbor)
                                sections.append(new_section)
                                new_section = [last_neighbor]
                                last_neighbor_location_nr = check_location_nr
                                last_neighbor = (check_x, check_y)
                            else:
                                last_neighbor = (check_x, check_y)

                        if input_matrix[check_x][check_y] == 1:
                            if start_x == check_x and start_y == check_y:
                                counter1 += 1

                                if new_check_location_nr == 1 or counter1 >= 3:
                                    inside = True
                                    break

                            check_location_nr = new_check_location_nr
                            new_x, new_y = check_x, check_y
                            counter2 = 0
                            matrix[check_x][check_y] = 1
                        else:
                            check_location_nr = 1 + (check_location_nr % 8)
                            if counter2 > 8:
                                counter2 = 0
                                break
                            else:
                                counter2 += 1

        return sections, matrix

    def __debug_showMatrix(self, array = None, awidth=None, aheight=None):
        array = self.__joint_matrix if array is None else array
        awidth = self.__m_width if awidth is None else awidth
        aheight = self.__m_height if aheight is None else aheight

        for y in reversed(range(aheight)):
            for x in range(awidth):
                print(array[x][y], end=" ")
            print()

        print()
