import copy
import numpy

from UM.Application import Application
from UM.Math.Vector import Vector
from UM.Mesh.MeshBuilder import MeshBuilder
from UM.Operations.AddSceneNodeOperation import AddSceneNodeOperation
from UM.Operations.GroupedOperation import GroupedOperation
from UM.Operations.RemoveSceneNodeOperation import RemoveSceneNodeOperation
from UM.Scene.SceneNode import SceneNode
from UM.Scene.Selection import Selection
from .Split import Split


def ModelSplitter():
    job = SplitModelsJob()
    job.start()


class SplitModelsJob(Split):
    def __init__(self):
        super().__init__()
        self._nodes_count = 0

    def run(self):
        selected_nodes = Selection.getCurrentlySelectedObjects()
        self._nodes_count = len(selected_nodes)
        self._initialize_progress(self._nodes_count)

        for node in selected_nodes:
            Application.getInstance().displayStatus("Splitting...")
            mesh_data = node.getMeshData()
            original_vertices = mesh_data.getVertices()
            vertices = copy.deepcopy(original_vertices)

            if mesh_data.hasIndices():
                faces = copy.deepcopy(node.getMeshData().getIndices())
            else:
                face_count = int(len(vertices)/3)
                faces = numpy.zeros((face_count, 3), dtype=numpy.int32)

                for i in range(face_count):
                    vertex_count = i*3
                    faces[i, 0] = vertex_count
                    faces[i, 1] = vertex_count + 1
                    faces[i, 2] = vertex_count + 2

            vertices = self._sort_vertices(vertices)
            vertex_to_face = self._map_vertices_to_faces(vertices, faces)
            separate_meshes = self._find_separate_meshes(vertex_to_face, faces, original_vertices)
            count = len(separate_meshes)
            if count > 1:
                self._replace_scene_node(node, separate_meshes)
                Application.getInstance().displayStatus("Splitted into " + str(count) + " meshes.")
            else:
                self._add_node_progress(0.1)
                Application.getInstance().displayStatus("Nothing to split.")

            self._finish_progress()

    def _sort_vertices(self, vertices):
        vertices = numpy.insert(vertices, 0, numpy.arange(len(vertices), dtype=numpy.int32), 1)
        vertices = vertices[vertices[:,3].argsort(kind='mergesort')]
        vertices = vertices[vertices[:,2].argsort(kind='mergesort')]
        vertices = vertices[vertices[:,1].argsort(kind='mergesort')]
        self._add_node_progress(0.1)

        return vertices

    def _map_vertices_to_faces(self, vertices, faces):
        unique_vertex = vertices[0]
        unique_vertex_id = int(unique_vertex[0])
        vertex_to_face = [[] for i in range(len(vertices))]

        signal_step = int(len(vertices) / 10)

        for i, vertex in enumerate(vertices):
            vertex_id = int(vertex[0])
            face_id = int(vertex_id/3)
            # Changes vertex id in faces array to unique vertex id when current vertex is copy of unique_vertex
            if vertex[1] == unique_vertex[1] and vertex[2] == unique_vertex[2] and vertex[3] == unique_vertex[3]:
                faces[face_id][vertex_id%3] = unique_vertex_id
            else:
                unique_vertex = vertex
                unique_vertex_id = int(unique_vertex[0])
            
            vertex_to_face[unique_vertex_id].append(face_id)

            if i % signal_step == 0:
                self._add_node_progress(1 / 10 * 0.3)

        return vertex_to_face

    def _find_separate_meshes(self, vertex_to_face, faces, original_vertices):
        done_set = set()
        separate_meshes = []

        for idx in range(0, len(faces)):
            if idx in done_set:
                continue
            done_set.add(idx)
            to_do_list = [idx]
            face_list = []
            while len(to_do_list) > 0:
                idx = to_do_list.pop()
                face_list.append(idx)
                for n in range(0, 3):
                    for i in vertex_to_face[faces[idx][n]]:
                        if not i in done_set:
                            done_set.add(i)
                            to_do_list.append(i)

            completed_faces_ratio = len(face_list)/len(faces)
            self._add_node_progress(completed_faces_ratio * 0.2)

            mesh_builder = MeshBuilder()
            mesh_builder.setFileName("Color it for me, please.")
            mesh_builder.reserveFaceCount(len(face_list))

            for i, face_id in enumerate(face_list):
                mesh_builder.addFaceByPoints(
                    original_vertices[faces[face_id, 0], 0], original_vertices[faces[face_id, 0], 1], original_vertices[faces[face_id, 0], 2],
                    original_vertices[faces[face_id, 1], 0], original_vertices[faces[face_id, 1], 1], original_vertices[faces[face_id, 1], 2],
                    original_vertices[faces[face_id, 2], 0], original_vertices[faces[face_id, 2], 1], original_vertices[faces[face_id, 2], 2])
                if i % 10000 == 0:
                    self._add_node_progress(10000 / len(faces) * 0.3)

            mesh_builder.calculateNormals(fast = True)
            mesh = mesh_builder.build()

            separate_meshes.append(mesh)

        return separate_meshes

    def _replace_scene_node(self, node, separate_meshes):
        node_transformation = node.getLocalTransformation()
        node_orientation = node.getOrientation()
        node_position = node.getPosition()
        separate_nodes = []

        Selection.clear()

        grouped_operation = GroupedOperation()

        for index, mesh in enumerate(separate_meshes):
            scene_node = SceneNode()
                        
            scene_node.setMeshData(mesh)
            scene_node.setSelectable(True)
            scene_node.setTransformation(copy.deepcopy(node_transformation))
            scene_node.setOrientation(copy.deepcopy(node_orientation))
            scene_node.setPosition(copy.deepcopy(node_position))

            center = Vector()
            for i in range(3):
                minimumum = numpy.amin(mesh._vertices[:,i])
                maximumum = numpy.amax(mesh._vertices[:,i])
                center._data[i] = maximumum - abs(maximumum - minimumum)/2

            scene_node.setCenterPosition(center)
            scene_node.translate(center, SceneNode.TransformSpace.World)
            scene_node.setName(self._get_new_node_name(node.getName(), index))

            scene = Application.getInstance().getController().getScene()

            operation = AddSceneNodeOperation(scene_node, scene.getRoot())
            grouped_operation.addOperation(operation)
            separate_nodes.append(scene_node)
            
        grouped_operation.addOperation(RemoveSceneNodeOperation(node))
        grouped_operation.push()

        if separate_nodes:
            Selection.clear()
            Selection.add(separate_nodes[0])
        
        self._add_node_progress(0.1)