// Copyright (c) 2015 Ultimaker B.V.
// Uranium is released under the terms of the AGPLv3 or higher.

import QtQuick 2.2
import QtQuick.Controls 1.1
import QtQuick.Controls.Styles 1.1
import QtQuick.Layouts 1.1

import UM 1.2 as UM
import Cura 1.0 as Cura

import Gence 1.0 as Gence

import "../../singletons"
import "../../buttons"

Item
{
    id: base

    property alias settingView: base
    
    property int tabWidth: width / 5
    property int tabHeight: Globals.fontPixelSize * 2.47

    property bool basicTabActive: true
    property bool printerTabActive: false
    property bool multipleExtrudersProfile: machineExtruderCount.properties.value > 1
    property bool scrollActive: contents.width < scrollView.width
    property int defaultActiveExtruder: 0

    signal categoryTabClicked(string tabName)

    Component.onCompleted: {
        widthFactor = 6.5
        heightFactor = 8.125
        loaderMarginsFactor = 0.01
        defaultActiveExtruder = ExtruderManager.activeExtruderIndex
        categoryTabClicked("Machine")
    }

    Row
    {
        id: mainTabs
        spacing: 1
        CustomTab {
            text: "Basic"

            checked: basicTabActive

            setWidth: tabWidth
            setHeight: tabHeight

            mouseArea.onClicked: {
                printerTabActive = false
                basicTabActive = true
            }
        }

        CustomTab {
            text: "Printer"

            checked: printerTabActive
            setWidth: tabWidth
            setHeight: tabHeight

            mouseArea.onClicked: {
                basicTabActive = false
                printerTabActive = true
            }
        }

        CustomTab {
            text: "1st Extruder"

            visible: multipleExtrudersProfile

            checked: ExtruderManager.activeExtruderIndex === 0 && !basicTabActive && !printerTabActive

            setWidth: tabWidth
            setHeight: tabHeight

            mouseArea.onClicked: {
                basicTabActive = false
                printerTabActive = false
                ExtruderManager.setActiveExtruderIndex(0)
            }
        } 

        CustomTab {
            text: "2nd Extruder"

            visible: multipleExtrudersProfile

            checked: ExtruderManager.activeExtruderIndex === 1  && !basicTabActive && !printerTabActive

            setWidth: tabWidth
            setHeight: tabHeight

            mouseArea.onClicked: {
                basicTabActive = false
                printerTabActive = false
                ExtruderManager.setActiveExtruderIndex(1)
            }
        }
    }

    Label {
        text: "Profile name"
        visible: basicTabActive
        height: Globals.controlHeight

        verticalAlignment: Text.AlignVCenter
        color: Colors.customSettingsLabel
        font.italic: true
        font.pixelSize: Globals.fontPixelSize
        font.family: Fonts.boxed

        anchors {
            left: parent.left
            leftMargin: Globals.margin
            rightMargin: Globals.margin
            right: profileName.left
            bottom: profileName.bottom
        }
    }

    Rectangle {
        id: profileName
        visible: basicTabActive

        width: (parent.width - Globals.margin) * 0.5

        anchors {
            rightMargin: Globals.margin
            right: parent.right
            top: mainTabs.bottom
            topMargin: Globals.margin
        }

        height: Globals.controlHeight

        border.width: 1
        border.color: Colors.customSettingsEditBorder

        TextInput
        {
            id: input

            clip: true

            anchors
            {
                left: parent.left
                leftMargin: Globals.margin
                right: parent.right
                rightMargin: scrollActive ? 1 : Globals.margin
                verticalCenter: parent.verticalCenter
            }

            text: Gence.CustomSettingsActions.profileName

            Keys.onReleased:
            {
                Gence.CustomSettingsActions.profileName = text
            }

            onEditingFinished:
            {
                Gence.CustomSettingsActions.profileName = text
            }

            color: Colors.customSettingsText
            selectionColor: Colors.customListSelectedBackground
            selectedTextColor: Colors.customListSelectedText
            font.family: Fonts.boxed
            font.pixelSize: Globals.fontPixelSize

            selectByMouse: true
        }
    }

    ListView {
        visible: basicTabActive
        anchors.bottom: basicWarning.top
        anchors.right: parent.right
        anchors.left: parent.left
        anchors.leftMargin: Globals.margin
        anchors.top: profileName.bottom
        anchors.topMargin: 1

        spacing: 1
        interactive: false

        model: ListModel
        {
            ListElement {
                label: "Machine type"
                key: "machine_name"
                type: "str"
                unit: ""
                showAlways: true
            }
            ListElement {
                label: "Support type"
                key: 'support_pattern'
                type: "enum"
                options: [
                    ListElement{ key: "lines"; value: "Lines" },
                    ListElement{ key: "grid"; value: "Grid" },
                    ListElement{ key: "triangles"; value: "Triangles" },
                    ListElement{ key: "concentric"; value: "Concentric" },
                    ListElement{ key: "concentric_3d"; value: "Concentric 3D" },
                    ListElement{ key: "zigzag"; value: "Zig Zag" }
                ]
                showAlways: true
            }
            ListElement {
                label: "Build adhesion type"
                key: "adhesion_type"
                type: "enum"
                options: [
                    ListElement{ key: "skirt"; value: "Skirt" },
                    ListElement{ key: "brim"; value: "Brim" },
                    ListElement{ key: "raft"; value: "Raft" },
                    ListElement{ key: "none"; value: "None" }
                ]
                showAlways: true
            }
            ListElement {
                label: "Flow"
                key: "material_flow"
                type: "float"
                unit: "%"
                extruder: -1
            }
            ListElement {
                label: "Flow T0"
                key: "material_flow"
                type: "float"
                unit: "%"
                extruder: 0
            }
            ListElement {
                label: "Flow T1"
                key: "material_flow"
                type: "float"
                unit: "%"
                extruder: 1
            }
            ListElement {
                label: "Infill density"
                key: "infill_sparse_density"
                type: "float"
                unit: "%"
                extruder: -1
            }
            ListElement {
                label: "Infill density T0"
                key: "infill_sparse_density"
                type: "float"
                unit: "%"
                extruder: 0
            }
            ListElement {
                label: "Infill density T1"
                key: "infill_sparse_density"
                type: "float"
                unit: "%"
                extruder: 1
            }
            ListElement {
                label: "Temperature*"
                key: "material_print_temperature"
                type: "float"
                unit: "°C"
                extruder: -1
            }
            ListElement {
                label: "Temperature T0*"
                key: "material_print_temperature"
                type: "float"
                unit: "°C"
                extruder: 0
            }
            ListElement {
                label: "Temperature T1*"
                key: "material_print_temperature"
                type: "float"
                unit: "°C"
                extruder: 1
            }
            ListElement {
                label: "Temperature heatbed*"
                key: "material_bed_temperature"
                type: "float"
                unit: "°C"
                showAlways: true
            }
            ListElement {
                label: "Cooling fan speed"
                key: "cool_fan_speed"
                type: "float"
                unit: "%"
                extruder: -1
            }
            ListElement {
                label: "Cooling fan speed T0"
                key: "cool_fan_speed"
                type: "float"
                unit: "%"
                extruder: 0
            }
            ListElement {
                label: "Cooling fan speed T1"
                key: "cool_fan_speed"
                type: "float"
                unit: "%"
                extruder: 1
            }
        }

        delegate: Loader
        {
            id: delegate

            width: parent.width
            height: visible ? Globals.controlHeight : -Globals.spacing
            visible: {
                if (model.showAlways){
                    return true
                } else if (model.extruder == -1 && machineExtruderCount.properties.value <= 1){
                    return true
                } else if (model.extruder != -1 && machineExtruderCount.properties.value > 1) {
                    return true
                }

                return false
            }
            enabled: visible

            property var definition: model
            property var settingDefinitionsModel: model
            property var propertyProvider: {
                if(model.showAlways || machineExtruderCount.properties.value <= 1){
                    return printerProvider
                } else if (model.extruder == 0) {
                    return ext0Provider
                } else if (model.extruder == 1) {
                    return ext1Provider
                }
                return printerProvider
            }
            property var globalPropertyProvider: inheritStackProvider

            //Qt5.4.2 and earlier has a bug where this causes a crash: https://bugreports.qt.io/browse/QTBUG-35989
            //In addition, while it works for 5.5 and higher, the ordering of the actual combo box drop down changes,
            //causing nasty issues when selecting different options. So disable asynchronous loading of enum type completely.
            //asynchronous: model.type != "enum" && model.type != "extruder"
            active: model.type != undefined

            source:
            {
                switch(model.type)
                {
                    case "int":
                        return "SettingTextField.qml"
                    case "float":
                        return "SettingTextField.qml"
                    case "enum":
                        return "SettingComboBox.qml"
                    case "extruder":
                        return "SettingExtruder.qml"
                    case "bool":
                        return "SettingCheckBox.qml"
                    case "str":
                        return "SettingTextField.qml"
                    case "polygons":
                        return "SettingTextArea.qml"
                    case "polygon":
                        return "SettingTextArea.qml"
                    case "category":
                        return "SettingCategory.qml"
                    default:
                        return "SettingUnknown.qml"
                }
            }

            Binding
            {
                target: printerProvider
                property: "containerStackId"
                when: model.settable_per_extruder || (inheritStackProvider.properties.limit_to_extruder != null && inheritStackProvider.properties.limit_to_extruder >= 0)
                value: {
                    //It's here only because of support pattern setting in dual extruder printer
                    if(machineExtruderCount.properties.value > 1 && inheritStackProvider.properties.limit_to_extruder != null && inheritStackProvider.properties.limit_to_extruder >= 0){
                        return ExtruderManager.extruderIds[String(inheritStackProvider.properties.limit_to_extruder)]
                    }

                    return Cura.MachineManager.activeMachineId
                }
            }

            UM.SettingPropertyProvider
            {
                id: printerProvider
                containerStackId: Cura.MachineManager.activeMachineId
                key: model.key ? model.key : ""
                watchedProperties: [ "value", "enabled", "state", "validationState", "settable_per_extruder", "resolve" ]
                storeIndex: 0
                // Due to the way setPropertyValue works, removeUnusedValue gives the correct output in case of resolve
                removeUnusedValue: model.resolve == undefined
            }

            UM.SettingPropertyProvider
            {
                id: ext0Provider
                containerStackId: machineExtruderCount.properties.value > 1 ? ExtruderManager.extruderStackId(0) : Cura.MachineManager.activeMachineId
                key: model.key ? model.key : ""
                watchedProperties: [ "value", "enabled", "state", "validationState", "settable_per_extruder", "resolve" ]
                storeIndex: 0
                // Due to the way setPropertyValue works, removeUnusedValue gives the correct output in case of resolve
                removeUnusedValue: model.resolve == undefined
            }

            UM.SettingPropertyProvider
            {
                id: ext1Provider
                containerStackId: machineExtruderCount.properties.value > 1 ? ExtruderManager.extruderStackId(1) : Cura.MachineManager.activeMachineId
                key: model.key ? model.key : ""
                watchedProperties: [ "value", "enabled", "state", "validationState", "settable_per_extruder", "resolve" ]
                storeIndex: 0
                // Due to the way setPropertyValue works, removeUnusedValue gives the correct output in case of resolve
                removeUnusedValue: model.resolve == undefined
            }

            // Specialty provider that only watches global_inherits (we cant filter on what property changed we get events
            // so we bypass that to make a dedicated provider).
            UM.SettingPropertyProvider
            {
                id: inheritStackProvider
                containerStackId: Cura.MachineManager.activeMachineId
                key: model.key
                watchedProperties: [ "limit_to_extruder" ]
            }
        }
    }

    Label {
        id: basicWarning
        visible: basicTabActive

        x: Globals.margin

        anchors.bottom: buttons.top
        anchors.bottomMargin: Globals.margin

        font.pixelSize: Globals.fontPixelSize
        font.family: Fonts.boxed
        color: Colors.customSettingsLabel

        text: "*Start GCode changes may be required"
    }

    GridView
    {
        id: tabs

        visible: !basicTabActive

        anchors.top: mainTabs.bottom
        anchors.topMargin: Globals.margin
        
        clip: true
        interactive: false

        cellWidth: tabWidth
        cellHeight: tabHeight

        width: parent.width
        height: Math.ceil(count/5) * tabHeight

        Component.onCompleted: model.update

        model: ListModel
        {
            ListElement {
                label: "Machine"
                key: "machine_settings"
            }
            ListElement {
                label: "Quality" 
                key: "resolution"
            }
            ListElement {
                label: "Shell"
                key: "shell"
            }
            ListElement {
                label: "Infill"
                key: "infill"
            }
            ListElement {
                label: "Material"
                key: "material"
            }
            ListElement {
                label: "Speed"
                key: "speed"
            }
            ListElement {
                label: "Travel"
                key: "travel"
            }
            ListElement {
                label: "Cooling"
                key: "cooling"
            }
            ListElement {
                label: "Support"
                key: "support"
            }
            ListElement {
                label: "Adhesion"
                key: "platform_adhesion"
            }
            ListElement {
                label: "Dual Extrusion"
                key: "dual"
            }
            ListElement {
                label: "Mesh Fixes"
                key: "meshfix"
            }
            ListElement {
                label: "Special Modes"
                key: "blackmagic"
            }
            ListElement {
                label: "Experimental"
                key: "experimental"
            }
        }

        delegate: Loader
        {
            id: delegate

            property var definition: model
            property var settingDefinitionsModel: definitionsModel

            width: parent.cellWidth

            source:
            {
                return "CategoryTab.qml"
            }
        }
    }

    ScrollView
    {
        id: scrollView
        visible: !basicTabActive
        anchors.top: tabs.bottom
        anchors.topMargin: Globals.margin
        anchors.bottom: buttons.top
        anchors.bottomMargin: Globals.margin
        anchors.right: parent.right
        anchors.left: parent.left

        Component.onCompleted: definitionsModel.setAllVisible(true)

        frameVisible: false

        style: ScrollViewStyle {

            decrementControl: null
            incrementControl: null

            scrollBarBackground: Item {
                implicitWidth: Globals.scrollBarWidth * 3
            }

            handle: Item {
                implicitWidth: Globals.scrollBarWidth * 3
                
                Rectangle {
                    width: Globals.scrollBarWidth
                    anchors {
                        top: parent.top
                        bottom: parent.bottom
                        horizontalCenter: parent.horizontalCenter
                    }
                    color: Colors.customListScrollBarHandle
                }
            }
        }

        ListView
        {
            id: contents
            spacing: Globals.spacing
            cacheBuffer: 1000000   // Set a large cache to effectively just cache every list item.

            boundsBehavior: ListView.StopAtBounds

            model: UM.SettingDefinitionsModel
            {
                id: definitionsModel
                containerId: Cura.MachineManager.activeDefinitionId
                exclude: ["command_line_settings", "infill_mesh", "infill_mesh_order", "support_mesh", "anti_overhang_mesh"] // TODO: infill_mesh settigns are excluded hardcoded, but should be based on the fact that settable_globally, settable_per_meshgroup and settable_per_extruder are false.
                visibilityHandler: UM.SettingPreferenceVisibilityHandler {}
                showAll: true
                onVisibilityChanged: Cura.SettingInheritanceManager.forceUpdate()
            }

            delegate: Loader
            {
                id: delegate

                width: parent.width
                visible: {
                    var globalSetting = !model.settable_per_extruder || model.limit_to_extruder != "-1"

                    if((machineExtruderCount.properties.value <= 1) || (printerTabActive && globalSetting) || (!printerTabActive && !globalSetting)){
                        return true
                    }
                    
                    return false
                }
                height: visible ? Globals.controlHeight : -Globals.spacing

                property bool editVisible: visible
                
                enabled:
                {
                    if(!ExtruderManager.activeExtruderStackId && ExtruderManager.extruderCount > 0)
                    {
                        // disable all controls on the global tab, except categories
                        return model.type == "category"
                    }
                    return visible
                }

                property var definition: model
                property var settingDefinitionsModel: definitionsModel
                property var propertyProvider: provider
                property var globalPropertyProvider: inheritStackProvider

                //Qt5.4.2 and earlier has a bug where this causes a crash: https://bugreports.qt.io/browse/QTBUG-35989
                //In addition, while it works for 5.5 and higher, the ordering of the actual combo box drop down changes,
                //causing nasty issues when selecting different options. So disable asynchronous loading of enum type completely.
                //asynchronous: model.type != "enum" && model.type != "extruder"
                active: model.type != undefined

                source:
                {
                    if (model.label === "End GCode" | model.label === "Start GCode") {
                        return "SettingTextArea.qml"
                    }
                    switch(model.type)
                    {
                        case "int":
                            return "SettingTextField.qml"
                        case "float":
                            return "SettingTextField.qml"
                        case "enum":
                            return "SettingComboBox.qml"
                        case "extruder":
                            return "SettingExtruder.qml"
                        case "bool":
                            return "SettingCheckBox.qml"
                        case "str":
                            return "SettingTextField.qml"
                        case "polygons":
                            return "SettingTextArea.qml"
                        case "polygon":
                            return "SettingTextArea.qml"
                        case "category":
                            return "SettingCategory.qml"
                        default:
                            return "SettingUnknown.qml"
                    }
                }

                // Binding to ensure that the right containerstack ID is set for the provider.
                // This ensures that if a setting has a limit_to_extruder id (for instance; Support speed points to the
                // extruder that actually prints the support, as that is the setting we need to use to calculate the value)
                Binding
                {
                    target: provider
                    property: "containerStackId"
                    when: model.settable_per_extruder || (inheritStackProvider.properties.limit_to_extruder != null && inheritStackProvider.properties.limit_to_extruder >= 0)
                    value:
                    {
                        if(!model.settable_per_extruder || machineExtruderCount.properties.value == 1)
                        {
                            //Not settable per extruder or there only is global, so we must pick global.
                            return Cura.MachineManager.activeMachineId
                        }
                        if(inheritStackProvider.properties.limit_to_extruder != null && inheritStackProvider.properties.limit_to_extruder >= 0)
                        {
                            //We have limit_to_extruder, so pick that stack.
                            return ExtruderManager.extruderIds[String(inheritStackProvider.properties.limit_to_extruder)]
                        }
                        if(ExtruderManager.activeExtruderStackId)
                        {
                            //We're on an extruder tab. Pick the current extruder.
                            return ExtruderManager.activeExtruderStackId
                        }
                        //No extruder tab is selected. Pick the global stack. Shouldn't happen any more since we removed the global tab.
                        return Cura.MachineManager.activeMachineId
                    }
                }

                // Specialty provider that only watches global_inherits (we cant filter on what property changed we get events
                // so we bypass that to make a dedicated provider).
                UM.SettingPropertyProvider
                {
                    id: inheritStackProvider
                    containerStackId: Cura.MachineManager.activeMachineId
                    key: model.key
                    watchedProperties: [ "limit_to_extruder" ]
                }

                UM.SettingPropertyProvider
                {
                    id: provider

                    containerStackId: Cura.MachineManager.activeMachineId
                    key: model.key ? model.key : ""
                    watchedProperties: [ "value", "enabled", "state", "validationState", "settable_per_extruder", "resolve" ]
                    storeIndex: 0
                    // Due to the way setPropertyValue works, removeUnusedValue gives the correct output in case of resolve
                    removeUnusedValue: model.resolve == undefined
                }
            }

            Menu
            {
                id: contextMenu

                property string key
                property var provider
                property bool settingVisible

                MenuItem
                {
                    text: "Copy value to all extruders"
                    visible: machineExtruderCount.properties.value > 1
                    enabled: contextMenu.provider != undefined && contextMenu.provider.properties.settable_per_extruder != "False"
                    onTriggered: Cura.MachineManager.copyValueToExtruders(contextMenu.key)
                }
            }

            UM.SettingPropertyProvider
            {
                id: machineExtruderCount

                containerStackId: Cura.MachineManager.activeMachineId
                key: "machine_extruder_count"
                watchedProperties: [ "value" ]
                storeIndex: 0
            }
        }
    }
    Grid {
        id: buttons
        anchors.bottom: parent.bottom
        width: parent.width / 2 - Globals.spacing / 2
        spacing: Globals.spacing
        columns: 2

        MenuButton {
            menuButtonText: "Load Profile"
            shadowEnabled: false
            buttonClickedAction.onTriggered: {
                Gence.CustomSettingsActions.loadProfileDialog()
            }
        }
        MenuButton {
            menuButtonText: "Save Profile"
            shadowEnabled: false
            buttonClickedAction.onTriggered: {
                Gence.CustomSettingsActions.exportProfileDialog()
            }
        }
        MenuButton {
            menuButtonText: "Activate Profile"
            shadowEnabled: false
            buttonClickedAction.onTriggered: {
                Gence.CustomSettingsActions.activateProfile()
                close()
            }
        }
        MenuButton {
            menuButtonText: "Discard and close"
            shadowEnabled: false
            buttonClickedAction.onTriggered: {
                close()
                Gence.CustomSettingsActions.loadLastProfile()
            }
        }
    }

    Connections {
        target: modalWindow
        onForcedClose: {
            Gence.CustomSettingsActions.loadLastProfile()
        }
        onClosed: {
            ExtruderManager.setActiveExtruderIndex(defaultActiveExtruder)
        }
    }
}